clc
close all
clear all

global mdl
global xdata
global ydata
global maxVarIdx

warning("off")
dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
filename = strcat(pwd,'\data4.xlsx');
delete("data4_*.txt");
outFile = strcat("data4_",dtstr,".txt");
diary(outFile)

fprintf("Program  clemTerm4\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ...\n\n");
fprintf("Excel file used is %s\n", filename);
fprintf("Diary file used is %s\n", outFile);
xyData = readmatrix(filename,'Sheet','Data');

ydata = xyData(:,1);
xdata = xyData(:,2:end);
[maxrows,maxcols] = size(xdata);
xyTransf = readmatrix(filename,'Sheet','Transf');
xyTransf = xyTransf(:,2:end);

% Important assignment - must match the number of X columns in sheet
% Transf of the Excel file
numX = 8;
Lb = xyTransf(1,:);
Ub = xyTransf(2,:);
% augment indices for variables to include the index of the dummy unity
% variable
Ub(1,1:numX) = Ub(1,1:numX)+1;
maxVarIdx = Ub(1,1:numX);
nVars = length(Lb);
iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx = Lb;
bStop = false;
iter = 0;
fprintf("------------- Calculating Max Iters ----------------\n");
while ~bStop 
  iter = iter +1;
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  % % % skip 0 .... can be deleted
  if idx(1) == 0, idx(1) = 1; end
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      % % % skip 0 .... can be deleted
      if idx(i) == 0, idx(i) = 1; end
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
str = num2str(iter);
str = fliplr(regexprep(fliplr(str), '(\d+\.)?(\d{3})(?=\S+)', '$1$2,'));
fprintf("Max iters = %s\n", str);
MaxIters = iter;
FivePercent = fix(iter/20);

iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx = Lb;
iter = 0;

bestX = zeros(1,length(Lb));
bestFx = 0;
% --------------------------------- start main loop
bStop = false;
iter = 0;
fprintf("------------- Maximizing Fx ----------------\n");
while ~bStop 
  iter = iter +1;
  if mod(iter, FivePercent) == 0
    percent = 100*iter/MaxIters;
    fprintf("Progress %5.2f%%\n", percent);
  end

  r2adj = myFit(idx);
  if r2adj > bestFx
    bestFx = r2adj;
    bestX = idx;
    bestMdl = mdl;
    fprintf("Iter %d, Fx = %e, X=[", iter, bestFx)
    fprintf("%d, ", bestX);
    fprintf("]\n");
    if r2adj == 1, break; end
  end
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  % % % skip 0 .... can be deleted
  if idx(i) == 0, idx(i) = 1; end
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      % % % skip 0 .... can be deleted
      if idx(i) == 0, idx(i) = 1; end
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
fprintf("Regression model\n")
format long
AdjR2 = bestMdl.Rsquared.Adjusted;
fprintf("Adjusted R-square = %14.12f\n", AdjR2);
mdl = bestMdl
bestX = [AdjR2 bestX];

res = zeros(2,length(bestX));
res(1,:) = bestX;
res(2,2:6) = mdl.Coefficients{1:5,1}';
writematrix(res(1,:), filename, 'Sheet', 'Results', 'Range', 'A2:Z2');
writematrix(res(2,2:6), filename, 'Sheet', 'Results', 'Range', 'B3:Z3');
 
% now do the % projections
ycalc = project(res);
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, filename, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));

smodel = sprintf("Y = %e + (%e)*", res(2,2), res(2,3));
idx = res(1,2:numX+1);
px = res(1,numX+2:2*numX+1);

if px(1)==0
  sX1 = sprintf("log(X%d)", idx(1));
elseif px(1)==4
  sX1 = sprintf("sqrt(X%d)", idx(1));
elseif px(1)==-4
  sX1 = sprintf("1/log(X%d)", idx(1));
elseif px(1)==1
  sX1 = sprintf("X%d", idx(1));
elseif px(1)>1
  sX1 = sprintf("X%d^%d", idx(1), px(1));
elseif px(1)==-1
  sX1 = sprintf("1/X%d", idx(1));
elseif px(1)<-1
  sX1 = sprintf("1/X%d^%d", idx(1), abs(px(1)));
else
  sX1 = sprintf("X%d", idx(1));
end

if px(2)==0
  sX2 = sprintf("log(X%d)", idx(2));
elseif px(2)==4
  sX2 = sprintf("sqrt(X%d)", idx(2));
elseif px(2)==-4
  sX2 = sprintf("1/log(X%d)", idx(2));
elseif px(2)==1
  sX2 = sprintf("X%d", idx(2));
elseif px(2)>1
  sX2 = sprintf("X%d^%d", idx(2), px(2));
elseif px(2)==-1
  sX2 = sprintf("1/X%d", idx(2));
elseif px(2)<-1
  sX2 = sprintf("1/X%d^%d", idx(2), abs(px(2)));
else
  sX2 = sprintf("X%d", idx(2));
end

if px(3)==0
  sX3 = sprintf("log(X%d)", idx(3));
elseif px(3)==4
  sX3 = sprintf("sqrt(X%d)", idx(3));
elseif px(3)==-4
  sX3 = sprintf("1/log(X%d)", idx(3));
elseif px(3)==1
  sX3 = sprintf("X%d", idx(3));
elseif px(3)>1
  sX3 = sprintf("X%d^%d", idx(3), px(3));
elseif px(3)==-1
  sX3 = sprintf("1/X%d", idx(3));
elseif px(3)<-1
  sX3 = sprintf("1/X%d^%d", idx(3), abs(px(3)));
else
  sX3 = sprintf("X%d", idx(3));
end

if px(4)==0
  sX4 = sprintf("log(X%d)", idx(4));
elseif px(4)==4
  sX4 = sprintf("sqrt(X%d)", idx(4));
elseif px(4)==1
  sX4 = sprintf("X%d", idx(4));
elseif px(4)==-4
  sX4 = sprintf("1/log(X%d)", idx(4));
elseif px(4)>1
  sX4 = sprintf("X%d^%d", idx(4), px(4));
elseif px(4)==-1
  sX4 = sprintf("1/X%d", idx(4));
elseif px(4)<-1
  sX4 = sprintf("1/X%d^%d", idx(4), abs(px(4)));
else
  sX4 = sprintf("X%d", idx(4));
end

if px(5)==0
  sX5 = sprintf("log(X%d)", idx(5));
elseif px(5)==4
  sX5 = sprintf("sqrt(X%d)", idx(5));
elseif px(5)==-4
  sX5 = sprintf("1/log(X%d)", idx(5));
elseif px(5)==1
  sX5 = sprintf("X%d", idx(5));
elseif px(5)>1
  sX5 = sprintf("X%d^%d", idx(5), px(5));
elseif px(5)==-1
  sX5 = sprintf("1/X%d", idx(5));
elseif px(5)<-1
  sX5 = sprintf("1/X%d^%d", idx(5), abs(px(5)));
else
  sX5 = sprintf("X%d", idx(5));
end

if px(6)==0
  sX6 = sprintf("log(X%d)", idx(6));
elseif px(6)==4
  sX6 = sprintf("sqrt(X%d)", idx(6));
elseif px(6)==-4
  sX6 = sprintf("1/log(X%d)", idx(6));
elseif px(6)==1
  sX6 = sprintf("X%d", idx(6));
elseif px(6)>1
  sX6 = sprintf("X%d^%d", idx(6), px(6));
elseif px(6)==-1
  sX6 = sprintf("1/X%d", idx(6));
elseif px(6)<-1
  sX6 = sprintf("1/X%d^%d", idx(6), abs(px(6)));
else
  sX6 = sprintf("X%d", idx(6));
end

if px(7)==0
  sX7 = sprintf("log(X%d)", idx(7));
elseif px(7)==4
  sX7 = sprintf("sqrt(X%d)", idx(7));
elseif px(7)==-4
  sX7 = sprintf("1/log(X%d)", idx(7));
elseif px(7)==1
  sX7 = sprintf("X%d", idx(7));
elseif px(7)>1
  sX7 = sprintf("X%d^%d", idx(7), px(7));
elseif px(7)==-1
  sX7 = sprintf("1/X%d", idx(7));
elseif px(7)<-1
  sX7 = sprintf("1/X%d^%d", idx(7), abs(px(7)));
else
  sX7 = sprintf("X%d", idx(7));
end

if px(8)==0
  sX8 = sprintf("log(X%d)", idx(8));
elseif px(8)==4
  sX8 = sprintf("sqrt(X%d)", idx(8));
elseif px(8)==-4
  sX8 = sprintf("1/log(X%d)", idx(8));
elseif px(8)==1
  sX8 = sprintf("X%d", idx(8));
elseif px(8)>1
  sX8 = sprintf("X%d^%d", idx(8), px(8));
elseif px(8)==-1
  sX8 = sprintf("1/X%d", idx(8));
elseif px(8)<-1
  sX8 = sprintf("1/X%d^%d", idx(8), abs(px(8)));
else
  sX8 = sprintf("X%d", idx(8));
end

if idx(1)<maxVarIdx(1) && idx(2)<maxVarIdx(2)
  s = sprintf("%s * %s", sX1, sX2);
  smodel = strcat(smodel, s);
elseif idx(1)<maxVarIdx(1)
  smodel = strcat(smodel, sX1);
elseif idx(2)<maxVarIdx(2)
  smodel = strcat(smodel, sX2);
end

smodel = sprintf("%s\n + (%e)*", smodel, res(2,4));

if idx(3)<maxVarIdx(3) && idx(4)<maxVarIdx(4)
  s = sprintf("%s * %s", sX3, sX4);
  smodel = strcat(smodel, s);
elseif idx(3)<maxVarIdx(3)
  smodel = strcat(smodel, sX3);
elseif idx(4)<maxVarIdx(4)
   smodel = strcat(smodel, sX4);
end

smodel = sprintf("%s\n + (%e)*", smodel, res(2,5));

if idx(5)<maxVarIdx(5) && idx(6)<maxVarIdx(6)
  s = sprintf("%s * %s", sX5, sX6);
  smodel = strcat(smodel, s);
elseif idx(5)<maxVarIdx(5)
  smodel = strcat(smodel, sX5);
elseif idx(6)<maxVarIdx(6)
  smodel = strcat(smodel, sX6);
end

smodel = sprintf("%s\n + (%e)*", smodel, res(2,6));

if idx(7)<maxVarIdx(7) && idx(8)<maxVarIdx(8)
  s = sprintf("%s * %s", sX7, sX8);
  smodel = strcat(smodel, s);
elseif idx(7)<maxVarIdx(7)
  smodel = strcat(smodel, sX7);
elseif idx(8)<maxVarIdx(8)
  smodel = strcat(smodel, sX8);
end
fprintf("%s\n", smodel)

format short
fprintf("Done!\n");
diary off
beep on
for i=1:3
  beep;
  pause(1);
end

function x = myfx(x,p)
  if p==0
    x = log(x);
  elseif p==4
    x = sqrt(x);  
  elseif p==-4
    x = 1./log(x);  
  elseif p > 1
    x = x.^p;
  elseif p < 0
    x = 1./x.^abs(p);
  end
end

function r = myFit(coeff)
  global mdl
  global xdata
  global ydata
  global maxVarIdx

  numX = length(maxVarIdx);
  [maxrows, maxcols] = size(xdata);
  y = ydata;
  idx = coeff(1:numX);
  px= coeff(numX+1:2*numX);
  x = zeros(maxrows,numX);
  x2 = zeros(maxrows,fix(numX/2));

  % check for redundant vars in a term
  if idx(1)==idx(2) || idx(3)==idx(4) || idx(5)==idx(6) || idx(7)==idx(8)
    r = 0;
    r = 0;
    return;
  end

  % check X1,X2 with X3,X4
  if idx(1)==idx(3) && px(1)==px(3) && ...
     idx(2)==idx(4) && px(2)==px(4) 
    r = 0;
    return;
  end
  
  % check X1,X2 with X3,X4
  if idx(1)==idx(4) && px(1)==px(4) && ...
     idx(2)==idx(3) && px(2)==px(3)
    r = 0;
    return;
  end

  % check X1,X2 with X5,X6
  if idx(1)==idx(5) && px(1)==px(5) && ...
     idx(2)==idx(6) && px(2)==px(6) 
    r = 0;
    return;
  end
  
  % check X1,X2 with X5,X6
  if idx(1)==idx(6) && px(1)==px(6) && ...
     idx(2)==idx(5) && px(2)==px(5) 
    r = 0;
    return;
  end

  % check X3,X4 with X5,X6
  if idx(3)==idx(5) && px(3)==px(5) && ...
     idx(4)==idx(6) && px(4)==px(6)  
    r = 0;
    return;
  end
  
  % check X3,X4 with X5,X6
  if idx(4)==idx(6) && px(4)==px(6) && ...
     idx(3)==idx(5) && px(3)==px(5) 
    r = 0;
    return;
  end

  % check X5,X6 with X7,X8
  if idx(5)==idx(7) && px(5)==px(7) && ...
     idx(6)==idx(8) && px(6)==px(8)  
    r = 0;
    return;
  end

  % check X5,X6 with X7,X8
  if idx(6)==idx(7) && px(6)==px(7) && ...
     idx(5)==idx(8) && px(5)==px(8)  
    r = 0;
    return;
  end

  for i=1:numX
    if idx(i)<maxVarIdx(i)
      x(:,i) = myfx(xdata(:,idx(i)),px(i));
    else
      x(:,i) = 1+zeros(maxrows,1); 
    end
  end

  i1=-1;
  i2=0;
  for i=1:fix(numX/2)
    i1=i1+2;
    i2=i2+2;
    if idx(i1)<maxVarIdx(i1) && idx(i2)<maxVarIdx(i2)
      x2(:,i) = x(:,i1).*x(:,i2);
    elseif idx(i2)<maxVarIdx(i2)
      x2(:,i) = x(:,i2);
    elseif idx(i1)<maxVarIdx(i1)
      x2(:,i) = x(:,i1);
    end
  end
  X = [x2(:,1) x2(:,2) x2(:,3) x2(:,4)];
  mdl = fitlm(X,y);
  r = mdl.Rsquared.Adjusted;
end

function ycalc = project(res)
  global xdata
  global maxVarIdx

  numX = length(maxVarIdx);
  [maxrows, maxcols] = size(xdata);

  idx = res(1,2:numX+1);
  px = res(1,numX+2:2*numX+1);
  x = zeros(maxrows,numX);
  
  for i=1:numX
    if idx(i)<maxVarIdx(i)
      x(:,i) = myfx(xdata(:,idx(i)),px(i)); 
    else
      x(:,i) = 1 + zeros(maxrows,1);
    end
  end

  ycalc = res(2,2) + res(2,3) * x(:,1).*x(:,2);
  ycalc = ycalc + res(2,4) * x(:,3).*x(:,4);
  ycalc = ycalc + res(2,5) * x(:,5).*x(:,6);
  ycalc = ycalc + res(2,6) * x(:,7).*x(:,8);
end